"""
Created on Mon Sep 16 13:27:44 2019

@author: oskar
@description: Oebb specific Mongodb Handler which inherits from the 'general' Mongodb Handler
"""


import json
import simplejson
import sys
import os
import jsonref

from copy import deepcopy
from pymongo import MongoClient
import pandas as pd
import numpy as np

sys.path.append(os.getcwd())
from libraries.log import Log
from libraries.configuration import default as cfg
from libraries.db_handlers.MongodbHandler import MongodbHandler
from libraries.Singleton_Threadsafe import SingletonThreadsafe


class OebbMongodbHandlerPool(metaclass=SingletonThreadsafe):
    '''
    '''

    def __init__(self, size: int = 10):
        self._size = size
        self._mongodb_handlers = [OebbMongodbHandler() for _ in range(size)]

    def aquire(self):
        while not self._mongodb_handlers:
            self._mongodb_handlers = [OebbMongodbHandler() for _ in range(self._size)]
            log.warning("Ran out of Mongodb handlers, 10 more have been added. Are you sure you've returned yours?")
        return self._mongodb_handlers.pop()
        
    def release(self, mongodb_handler):
        if len(self._mongodb_handlers) < self._size:
            self._mongodb_handlers.append(mongodb_handler)


class OebbMongodbHandler(MongodbHandler):

    def __init__(self, database_url: str = None,
                 database_name: str = None):
        '''
        '''
        super().__init__(
                database_url=database_url,
                database_name=database_name)

        self._log = Log("Oebb Mongodb Handler")


    def query_process_instances_with_stage_filtering(self, stage_path: list):
        '''
        :param list stage_path: Stage path for which the data will be queried. 

        Method will return all wheelsets which stage_path is a subset of their final stage path. 
        The wheelsets will only be returned with the actual stations from the stages in the stage_path.
        '''
    
        assert(isinstance(stage_path, list)),\
            "Parameter 'stage_path must be a list type"

        aggregation_pipeline = [
                                {"$match": {"$expr": {"$setIsSubset": [stage_path, '$final_state.stage_path']}}},
                                {"$unwind": "$process"},
                                {"$sort": {"process.step_number": 1}},
                                {"$group": {
                                    "_id": {
                                            "_id": "$_id", 
                                            "radsatznummer": "$radsatznummer",
                                            "final_state": "$final_state"
                                            }, 
                                        "process": {"$push": "$process"}
                                    }
                                },
                                {"$project": {
                                    "_id": "$_id._id",
                                    "radsatznummer": "$_id.radsatznummer",
                                    "final_state": "$_id.final_state",
                                    "process":{
                                        "$filter":{
                                            "input": "$process",
                                            "as": "process",
                                            "cond":{
                                                "$setIsSubset": [["$$process.stage"], stage_path]
                                            },      
                                        }
                                    }
                                }
                                }
                            ]

        return self.aggregate_data_and_generate_dataframe('process_instances',aggregation_pipeline)                

    def query_process_instances_with_guaranteed_station_sorting(self, attribute: str = None, 
                                                attribute_value: str = None, comparison_operator: str = '$eq', ascending: bool = True):
        '''
        :param str attribute:
        :param str attribute_value:
        :param str comparison_operator:
        :param bool ascending:

        Method will return all wheelsets which stage_path is a subset of their final stage path. 
        The wheelsets will only be returned with the actual stations from the stages in the stage_path.
        '''
        if attribute is not None:
            assert(isinstance(attribute, str)),\
                "Parameter 'attribute' must be a string type"

        if attribute_value is not None:
            assert(isinstance(attribute_value, (str, int, bool))),\
                "Parameter 'attribute_value' must be a string, integer or boolean type"
    
        assert(isinstance(comparison_operator, str)),\
            "Parameter 'comparison_operator' must be a string type"

        assert(isinstance(ascending, bool)),\
            "Parameter 'ascending' must be a boolean type"

        
        order = 1 if ascending else -1

        aggregation_pipeline = [
                                {"$match":{ attribute:{comparison_operator: attribute_value}}},
                                {"$unwind": "$process"},
                                {"$sort": {"process.step_number": order}},
                                {"$group": {
                                    "_id": {
                                            "_id": "$_id", 
                                            "radsatznummer": "$radsatznummer",
                                            "final_state": "$final_state"
                                            }, 
                                        "process": {"$push": "$process"}
                                    }
                                },
                                {"$project": {
                                    "_id": "$_id._id",
                                    "radsatznummer": "$_id.radsatznummer",
                                    "final_state": "$_id.final_state",
                                    "process": "$process"}
                                }
                            ]
                    
        return self.aggregate_data_and_generate_dataframe('process_instances',aggregation_pipeline)